from collections import namedtuple
import random as rd
import math

Color=namedtuple('Color',['r','g','b'])

BLUE       = Color(  0,  0,255)
RED        = Color(255,  0,  0)
GREEN      = Color(  0,255,  0)
YELLOW     = Color(255,255,  0)
BLACK      = Color(  0,  0,  0)
WHITE      = Color(255,255,255)
GREY       = Color(100,100,100)
PURPLE     = Color(100,  0,100)
ORANGE     = Color(255,165,  0)
BROWN      = Color(160, 82, 45)
HALFGREY   = Color( 50, 50, 50)
DARKGREY   = Color( 20, 20, 20)
DARKRED    = Color( 10, 10, 10)
DARKGREEN  = Color(  0,100,  0)
DARKBLUE   = Color(  0,  0,100)
LIGHTRED   = Color(255,200,200)
LIGHTGREEN = Color(200,255,200)
LIGHTBLUE  = Color(200,200,255)
LIGHTBROWN = Color(229,219,222)
LIGHTGREY  = Color(200,200,200)
BEIGE      = Color(199,175,138)

sigmoid = s = lambda x:1/(1+math.exp(-x))
reverse_sigmoid = lambda x:math.log(x/(1-x))
#r=reverse_sigmoid()

bijection = lambda x,e,s:(x-e[0])/(e[1]-e[0])*(s[1]-s[0])+s[0]

random    = lambda :            Color(*[rd.randint(0,255)              for i in range(3)])
reverse   = lambda color:       Color(*[255-c                          for c in color])
darken    = lambda color,n=0:   Color(*[int(c*sigmoid(n/10))           for c in color])
lighten   = lambda color,n=0:   Color(*[int(255-(255-c)*sigmoid(n/10)) for c in color])
mix       = lambda cl1,cl2:     Color(*[(c1+c2)//2                     for (c1,c2) in zip(cl1,cl2)])
substract = lambda cl1,cl2:     Color(*[max(min(2*c1-c2,255),0)        for (c1,c2) in zip(cl1,cl2)])
increase  = lambda color,n=2:   Color(*[int(255*math.exp(n*math.log(c/255)))     for c in color])


def nuance(color1,color2,t,p=1/2):
    """Return a color between the two depending on the degree."""
    return Color(*[c1*(1-t**p)+c2*(t**p) for (c1,c2) in zip(color1,color2)])


def nuances(colors,t):
    """Does an interpolation between all the colors in order."""
    n=len(colors)
    int(t*n)

def ponderNuances(colors,t):
    """Does an interpolation between all the colors in order."""
    n=len(colors)
    int(t*n)





def getFromWavelength(wavelength):
    """Return a color using wavelength."""
    gamma,max_intensity=0.80,255
    def adjust(color, factor):
        if color==0: return 0
        else: return round(max_intensity*pow(color*factor,gamma))
    if 380<=wavelength<=440: r,g,b=-(wavelength-440)/(440-380),0,1
    elif 440<=wavelength<=490: r,g,b=0,(wavelength-440)/(490-440),1
    elif 490<=wavelength<=510: r,g,b=0,1,-(wavelength-510)/(510-490)
    elif 510<=wavelength<=580: r,g,b=(wavelength-510)/(580-510),1,0
    elif 580<=wavelength<=645: r,g,b=1,-(wavelength-645)/(645-580),0
    elif 645<=wavelength<=780: r,g,b=1,0,0
    else: r,g,b=0,0,0
    if 380<=wavelength<=420: factor=0.3+0.7*(wavelength-380)/(420-380)
    elif 420<=wavelength<=701: factor=1
    elif 701<=wavelength<=780: factor=0.3+0.7*(780-wavelength)/(780-700)
    else: factor=0
    r,g,b=adjust(r,factor),adjust(g,factor),adjust(b,factor)
    return Color(r,g,b)

def colorRange(t):
    bt=bijection(t,[0,1],[380,780])
    return getFromWavelength(bt)





if __name__=="__main__":
    print(darken(RED,10))
    print(mix(YELLOW,RED))
    print(reverse(LIGHTBROWN))
    print(substract(LIGHTBROWN,ORANGE))
    print(increase(LIGHTBROWN))
    print(nuance(YELLOW,RED,10))
    print("colorRange(0.5):",colorRange(0.1))
    print(BLUE.r)
    print("mycolors imported")



from .interpolation import PolynomialInterpolation
from .abstract import Point, Vector, Circle
from .curves import Trajectory
from pygame.locals import *
from . import colors

import numpy as np
import pygame
import pickle
import cmath
import math
import cv2
import sys
import os

class Fourier:

    def transform(pts, ncfs, wo=2 * math.pi):
        """Apply the true fourier transform by
        returning a dictionary of the coefficients."""
        npts = len(pts)
        h = ncfs // 2
        cfs = {}
        # Compute all coefficients
        for n in range(-h, h + 1):
            # Compute each coefficient
            cn = 0
            for iw in range(npts):
                w = iw / npts  # w is not a frequency but the variable of a parametric equation
                fw = complex(*pts[iw])
                cn += fw * cmath.exp(-1j * n * w * wo)
            cn /= npts
            cfs[n] = cn
        return cfs

    def inverseTransform(cfs, npts, wo=2 * math.pi):
        """Apply the true fourier inverse transform
        by returning the list of the points."""
        ncfs = len(cfs)
        h = npts // 2
        pts = []
        # Compute all the points
        for it in range(npts):
            t = it / npts  # t is not a time but the variable of a parametric equation of the final graph
            # Compute each point
            zpt = 0
            for (n, cn) in cfs.items():  # Addition is commutative, even though the dictionary is unordered, the sum of the terms will be the same
                zpt += cn * cmath.exp(1j * wo * n * t)
            pts.append((zpt.real, zpt.imag))
        return pts

    def build(cfs, t, wo=2 * math.pi):
        """Return the 'construction graph' with a given time 't'."""
        ncfs = len(cfs)
        h = ncfs // 2
        cst = [(0, 0)]
        zpt = cfs[0]
        cst.append((zpt.real, zpt.imag))
        for n in range(1, h + 1):
            pcf = cfs[n] * cmath.exp(1j * wo * n * t)
            ncf = cfs[-n] * cmath.exp(1j * wo * (-n) * t)
            zpt += pcf
            cst.append((zpt.real, zpt.imag))
            zpt += ncf
            cst.append((zpt.real, zpt.imag))
        return cst


class VisualFourier:
    """Show an application of the fourier transform."""

    # Instance methods
    def __init__(self,
                 context,
                 image=None,
                 coefficients=[],
                 directory="../FourierObjects",
                 filename="Fourier",
                 coefficients_filename="fourier_coefficients.txt"
                 ):
        """Initialization."""
        self.context = context
        self.coefficients = coefficients
        self.coefficients_filename = coefficients_filename

        # Directory
        self.directory = directory
        self.filename = filename

        # Graphs
        self.graphs = [[], [], []]

        # Memory cache for less calculations
        # This information is redundant and can be computed again
        self.sample = []

        # Mode
        self.mode = 0
        self.step = 0
        self.max_step = 1000
        self.messages = ['drawing', 'construction', 'display']
        self.pause = False
        self.include = True  # Include the last point of the interpolation

        # Precision settings
        # self.coefficients_number=100
        self.sample_number = 5
        self.display_number = 100  # Number of points of the display graph
        self.integral_precision = 100
        self.points_radius = 5

        # Optional settings
        # Graph shown
        self.show_image = True
        self.show_polynomial = False
        self.show_drawing = True
        self.show_display = True
        self.show_vectors = True
        self.show_circles = True
        self.show_sample = True
        self.show_camera = False
        # Graph color
        self.color_polynomial = colors.BLUE
        self.color_drawing = colors.GREEN
        self.color_display = colors.RED
        self.color_vectors = colors.WHITE
        self.color_circles = colors.GREY
        self.color_sample = colors.YELLOW

        # Set the image for sampling
        if image is None:
            self.image = None
            self.show_image = False
        else:
            self.image = self.context.loadImage(image)
            # Trying to convert a pygame image into a pil image that can be used for the canny algorithm.
            # s=self.image.get_size()
            # self.image=pygame.image.tostring(self.image,"RGBA",False)
            # self.image=Image.frombytes("RGBA",s,self.image)
            # self.image=cv2.Canny(self.image,50,150)

    def __call__(self):
        """Main loop."""
        self.setMode(self.mode)
        while self.context.open:
            self.events()
            self.main()
            self.show()

    def events(self):
        """Deal with all the events."""
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.context.open = False
            if event.type == KEYDOWN:

                if event.key == K_ESCAPE:
                    self.context.open = False
                if event.key == K_SPACE or event.key == K_MENU or event.key == K_q:
                    self.setMode((self.mode + 1) % 3)
                if event.key == K_0:
                    self.show_polynomial = not(self.show_polynomial)
                if event.key == K_1:
                    self.show_image = not(self.show_image)
                if event.key == K_2:
                    self.show_drawing = not(self.show_drawing)
                if event.key == K_3:
                    self.show_display = not(self.show_display)
                if event.key == K_4:
                    self.show_vectors = not(self.show_vectors)
                if event.key == K_5:
                    self.show_circles = not(self.show_circles)
                if event.key == K_6:
                    self.show_sample = not(self.show_sample)
                if event.key == K_r:
                    self.reset()
                if event.key == K_z:
                    self.drawing = self.drawing[:-1]
                    self.updateSample()
                if event.key == K_s:
                    self.save()  # Save the coefficients and the graphs
                if event.key == K_d:
                    self.saveCoefficients()
                if event.key == K_a:
                    # Save a picture the screen
                    self.screenshot(self.directory)
                if event.key == K_p:
                    self.pause = not(self.pause)
                if event.key == K_f:
                    self.context.switch()
                if event.key == K_c:
                    self.show_camera = not(self.show_camera)
                    if self.show_camera:
                        self.context.camera.buildCapture()
                    else:
                        self.context.camera.destroy()
            if event.type == pygame.MOUSEBUTTONDOWN:
                if (event.button == 1) and (self.mode == 0):
                    self.place()
                    self.updateSample()
                if event.button == 4:
                    self.context.draw.plane.zoom([1.1, 1.1])
                if event.button == 5:
                    self.context.draw.plane.zoom([0.9, 0.9])

            if event.type == VIDEORESIZE:
                self.context.screen = pygame.display.set_mode(
                    (event.w, event.h), RESIZABLE)

    def main(self):
        """Code inside the loop."""
        if self.mode == 0:  # drawing
            pass
        elif self.mode == 1:  # construction
            if self.step > self.max_step:
                self.mode = 2
            else:
                self.construction = Fourier.build(self.coefficients, self.time)
                self.display.append(self.construction[-1])
                if not self.pause:
                    self.step += 1
        elif self.mode == 2:  # display
            pass

    def show(self):
        """Show the graph."""
        self.context.control()
        self.context.clear()
        self.context.show()

        sx, sy = self.context.size
        drawing, construction, display = range(3)

        if self.show_image and self.image:
            w, h = self.image.get_size()
            r = h/w
            self.context.draw.image(self.context.screen, self.image, (-1/2, r*1/2))
        if self.show_camera:
            self.context.camera.show()

        if self.mode == 0:
            if self.show_polynomial:
                self.drawPolynomial(drawing, self.color_polynomial)
            if self.show_drawing:
                self.drawGraph(drawing, self.color_drawing)
            if self.show_sample:
                self.drawPoints(self.sample, self.color_sample)
        elif self.mode == 1:
            if self.show_polynomial:
                self.drawPolynomial(drawing, self.color_polynomial)
            if self.show_drawing:
                self.drawGraph(drawing, self.color_drawing)
            if self.show_vectors:
                self.drawVectors(construction, self.color_vectors)
            if self.show_circles:
                self.drawCircles(construction, self.color_circles)
            if self.show_display:
                self.drawGraph(display, self.color_display)
            if self.show_sample:
                self.drawPoints(self.sample, self.color_sample)
        elif self.mode == 2:
            if self.show_polynomial:
                self.drawPolynomial(drawing, self.color_polynomial)
            if self.show_drawing:
                self.drawGraph(drawing, self.color_drawing)
            if self.show_display:
                self.drawGraph(display, self.color_display)
            if self.show_sample:
                self.drawPoints(self.sample, self.color_sample)

        self.context.print("Time: " + str(self.time),
                           position=(10, 10), size=35, conversion=False)
        if self.pause:
            self.context.print("Pause", position=(
                sx - 100, 10), size=35, conversion=False)
        self.context.showConsole()
        self.context.flip()

    def reset(self):
        """Reset the graphs, the sample and the mode."""
        self.mode = 0
        self.graphs = [[], [], []]
        self.coefficients = []
        self.sample = []

    def getVectors(self, graph):
        """Return the list of vectors."""
        return [Vector.createFromTwoTuples(graph[i], graph[i + 1]) for i in range(len(graph) - 1)]

    def showVectors(self, graph, vectors, color=colors.WHITE):
        """Show the vectors on the screen."""
        for i in range(len(vectors) - 1):
            vectors[i].show(self.context, Point(*graph[i]), color)

    def setMode(self, mode):
        """Change the mode into another."""
        self.mode = mode
        if self.mode == 0:
            self.setDrawingMode()
        elif self.mode == 1:
            self.setConstructionMode()
        elif self.mode == 2:
            self.setDisplayMode()
        self.context.text.append("mode: " + self.messages[self.mode])

    def setDrawingMode(self):
        """Set the attributes before starting the drawing mode."""
        pass

    def setConstructionMode(self):
        """Set the attributes before starting the construction mode."""
        self.step = 0
        self.display = []
        # t=Trajectory.createFromTuples(self.drawing)
        # l=t.sampleSegments(self.sample_number)
        self.coefficients = Fourier.transform(
            self.sample, self.coefficients_number)

    def setDisplayMode(self):
        """Set the attributes before starting the display mode."""
        self.step = (self.max_step + int(self.include))
        self.display = Fourier.inverseTransform(
            self.coefficients, self.display_number)

    def place(self):
        """Place a point."""
        p = self.context.point()
        self.drawing.append(p)

    def updateSample(self):
        """Update the sample."""
        t = Trajectory.createFromTuples(self.drawing)
        self.sample = t.sampleSegments(
            self.sample_number, include=self.include)

    def screenshot(self):
        """Make a screenshot of the window."""
        self.context.draw.window.screenshot(self.filename)

    @property
    def dictionary(self):
        """Return the dictionary."""
        dictionary = {
            "coefficients":     self.coefficients,
            "drawing":          self.drawing,
            "construction":     self.construction,
            "display":          self.display
        }
        return dictionary

    def save(self):
        """Save the sampled graph and fourier's coefficients."""
        path = self.directory + "/" + self.filename
        pickle.dump(self.dictionary, open(path, 'wb'))
        self.context.console.append("The Fourier components are saved.")

    def saveCoefficients(self):
        """Save the coefficients in a txt file."""
        path = self.directory + "/" + self.coefficients_filename
        with open(path, mode="w", encoding="utf-8") as file:
            file.write(
                "\n".join([f"{k}:{v}" for k, v in self.dictionary["coefficients"].items()]))
            self.context.console.append(
                "The Fourier coefficients are written.")

    def load(self):
        """Load the fourier's coefficients."""
        path = self.directory + "/" + self.filename
        dictionary = pickle.load(open(path, 'rb'))
        self.coefficients = dictionary["coefficients"]
        self.display = dictionary["display"]
        self.construction = dictionary["construction"]
        self.drawing = dictionary["drawing"]
        self.updateSample()
        self.context.console.append("The Fourier components are loaded.")

    def getTime(self):
        """Return the time of the construction."""
        return self.step / (self.max_step + int(self.include))

    time = property(getTime)

    # Graphical functions
    def distance(self, p1, p2):
        """Return the distance between 2 points."""
        return math.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)

    def drawVectors(self, index, color):
        """Draw the vectors from the points."""
        graph = self.graphs[index]
        for i in range(len(graph) - 1):
            v = Vector.createFromTwoTuples(graph[i], graph[i + 1], color=color)
            v.showFromTuple(self.context, graph[i])

    def drawCircles(self, index, color):
        """Draw the circles from the points."""
        graph = self.graphs[index]
        for i in range(len(graph) - 1):
            radius = self.distance(graph[i], graph[i + 1])
            c = Circle.createFromPointAndRadius(graph[i], radius, color=color)
            c.show(self.context)

    def drawGraph(self, index, color, connected=False, width=1, conversion=True):
        """Draw the graph."""
        graph = self.graphs[index]
        if len(graph) > 1:
            self.context.draw.lines(
                self.context.screen, color, graph, connected, width, conversion)

    def drawPolynomial(self, index, color, precision=200):
        """Draw the polynomial interpolation of the points."""
        graph = self.graphs[index]
        if len(graph) > 1:
            p = PolynomialInterpolation(graph, color)
            p.show(self.context, precision)

    def drawPoints(self, points, color):
        """Draw the points that were sampled for the fourier transform."""
        for p in points:
            Point
            p.color = color
            p.radius = self.points_radius
            p.conversion = False
            p.show(self.context)

    def drawSample(self, index, color):
        """Draw the points that were sampled for the fourier transform."""
        t = Trajectory.createFromTuples(self.graphs[index])
        l = t.sampleSegments(self.sample_number, include=self.include)
        for e in l:
            p = Point(*e, radius=5, conversion=False)
            p.show(self.context)

    # Properties

    def getDrawing(self):
        return self.graphs[0]

    def setDrawing(self, graph):
        self.graphs[0] = graph

    def getConstruction(self):
        return self.graphs[1]

    def setConstruction(self, graph):
        self.graphs[1] = graph[1:]

    def getDisplay(self):
        return self.graphs[2]

    def setDisplay(self, graph):
        self.graphs[2] = graph

    def getCoefficientsNumber(self):
        return len(self.sample)

    drawing = property(getDrawing, setDrawing)
    construction = property(getConstruction, setConstruction)
    display = property(getDisplay, setDisplay)
    coefficients_number = property(getCoefficientsNumber)


if __name__ == "__main__":
    from context import Context
    import logging

    folder = "../FourierImages"

    if len(sys.argv) > 1:
        image_name = sys.argv[1]
    else:
        logging.warning("You must place your image in the FourierImages folder before using it.""")
        image_name = input('image name:')

    image = os.path.join(folder, image_name)
    print(image)


    context = Context(name="Application of the Fourier Transform.", fullscreen=False)
    fourier = VisualFourier(context, image=image, directory="../FourierObjects")
    fourier.load()
    fourier()
    fourier.save()



"""
The goal of this program is to make an interpolation of the points of any
curve in order to make it a parametric function.


There are 2 main types of interpolations:
- Polynomial Interpolation:
- Bezier Interpolation:
"""

#Polynomial Interpolation
from .polynomial import Polynomial

#Pseudo Bezier Interpolation
from .curves import BezierCurve
from .abstract import Line,Point
#additional imports for showing the interpolations more easily
from .curves import Trajectory
from . import colors

#Deprectated function
def directInterpolation(points,t):
    x=[pt[0] for pt in points]
    y=[pt[1] for pt in points]
    plx=Polynomial.createFromInterpolation(range(len(x)),x)
    ply=Polynomial.createFromInterpolation(range(len(y)),y)
    l=len(pts)
    mst=100

    # Creation of the points
    # pts=[(random.uniform(xmin,xmax),random.uniform(ymin,ymax)) for i in range(n)]

    # # """ Tests:
    # # Calculations for n points in dimension 2:
    #
    # x=[pt[0] for pt in pts]
    # y=[pt[1] for pt in pts]
    # plx=Polynomial.createFromInterpolation(range(len(x)),x)
    # ply=Polynomial.createFromInterpolation(range(len(y)),y)
    # print(plx)
    # print(ply)
    # l=len(pts)
    # mst=100
    # npts=[(plx(l*st/mst),ply(l*st/mst)) for st in range(mst)]
    # grapher=Grapher(context,functions=[plx,ply])
    # """


class Interpolation:
    """Base class of all interpolations."""

    def __init__(self,points,create=True):
        """Create the interpolation using a list of n-dimensional points."""
        self.points=points
        if create: self.create()

    def create(self):
        """Create the necessary attributes for the interpolation.
        This method is to be overloaded."""
        pass

    def sample(self,n):
        """Make a sample of the interpolation for n points."""
        m=n-1
        return [self(t/m) for t in range(n)]

    def __str__(self):
        """Return the string representation of the interpolation."""
        return "Itp("+",".join(map(str,self.points))+")"


class ExtraInterpolation(Interpolation):
    """Extra features for interpolations."""

    def getSegments(self):
        """Return the segments formed by the points."""
        return [Segment(self.points[i],self.points[i+1]) for i in range(len(self.points)-1)]

    def getTrajectory(self):
        """Return the trajectory associated with the points of the interpolation."""
        return Trajectory.createFromTuples(self.points)

    def getPointsType(self):
        """Return the points under a point type instead of just a tuple type."""
        return [Point(*p) for p in self.points]

    def getPoint(self,t,**kwargs):
        """Return the point of position t in the parametric function."""
        return Point(*self(t),**kwargs)

    segments=property(getSegments)
    trajectory=property(getTrajectory)
    points_type=property(getPointsType)


class VisualInterpolation(Interpolation):
    """Visual features for interpolation, this class allows to show an
    interpolation on a graphical context."""

    def __init__(self,points,
                    color=colors.RED,
                    point_color=colors.LIGHTRED,
                    trajectory_color=colors.GREEN,
                    **kwargs):
        """Create a visual interpolation."""
        self.color=color
        self.point_color=point_color
        self.trajectory_color=trajectory_color
        Interpolation.__init__(self,points,**kwargs)

    def show(self,context,n,color=None):
        """Show the polynomial interpolation by sampling the function using:
        -n: the number of points
        -context:the context in which the interpolation is shown."""
        if not color: color=self.color
        points=self.sample(n)
        context.draw.lines(context.screen,color,points,width=1,connected=False)

    def showTrajectory(self,context,color=None):
        """Show the trajectory defined by the points of the interpolation."""
        if not color: color=self.trajectory_color
        Trajectory.createFromTuples(self.points).show(context,color=color)


class PolynomialInterpolation(VisualInterpolation,ExtraInterpolation):
    """Make an interpolation using the lagrangian polynomial interpolator
    over list of n-dimensional points."""

    def create(self):
        """Create the components useful for the interpolation which are the
        split components and the lagrangian polynomials."""
        self.createComponents()
        self.createPolynomials()

    def createComponents(self):
        """Split its components of each point in separate lists."""
        self.components=[[pt[i] for pt in self.points] for i in range(len(self.points[0]))]

    def createPolynomials(self):
        """Create its polynomials using its components."""
        lpts=len(self.points)
        self.polynomials=[Polynomial.createFromInterpolation(range(lpts),c) for c in self.components]

    def __call__(self,t):
        """Evaluate the parametric interpolation for the input 't' between 0 and 1."""
        lcps=len(self.components)
        lpts=len(self.points)-1
        return tuple([self.polynomials[i](lpts*t) for i in range(lcps)])

class PseudoBezierInterpolation(VisualInterpolation):
    """Make an interpolation using the bezier interpolation over a list of
    n-dimensional points."""

    def create(self):
        """Create the lines, the intruding points and the bezier curves."""
        self.createLines()
        self.createIntrudingPoints()
        self.createBezierCurves()

    def createLines(self):
        """Create the lines of the duos of consecutive points."""
        l=len(self.points)
        lines=[Line.createFromTwoPoints(self.points[i],self.points[i+1]) for i in range(l-1)]

    def createIntrudingPoints(self):
        """Create the points of intersection formed by some of the lines."""
        l=len(lines)
        intruding_points=[lines[i].crossLine(lines[i+1]) for i in range(l-1)]

    def __call__(self,t):
        """Evaluate the parametric interpolation for the input 't' between 0 and 1."""
        l=len(lines)
        intruding_points=[lines[i].crossLine(lines[i+1]) for i in range(l-1)]
        bezier_curves=[]
        for i in range(l):
            bezier_points=[self.points[i],self.intruding_points[i],self.points[i+1]]
            b=BezierCurve(bezier_points)
            bezier_curves.append(b)



if __name__=="__main__":
    #To show the points
    from mycontext import Context
    import colors
    #To create the points
    import random

    context=Context(name="Interpolation Demonstration")

    #Parameters
    h=1; w=1; xmin,xmax,ymin,ymax=[-w,w,-h,h] #corners
    l=5 #number of points
    n=0 #number of steps
    m=100 #max number of steps

    #Final version
    #pts=[Point.random() for i in range(10)]
    pts=[Point(2*x,random.randint(-5,5)) for x in range(l)]
    interpolation=PolynomialInterpolation([p.components for p in pts])
    interpolation.color
    npts=interpolation.sample(200)

    #Setting the console
    context.console.append(["interpolation: "+str(interpolation),"trajectory: "+str(interpolation.trajectory)])


    #Main loop of the context
    while context.open:
        #Update the context
        context.check()
        context.control()
        context.clear()
        context.show()

        #Update data components
        Point.turnPoints([1/100/l for i in range(l)],pts)
        #trajectory=Trajectory(pts,colors.GREEN)
        interpolation=PolynomialInterpolation([p.components for p in pts])
        npts=interpolation.sample(200) #Sample 200 points by interpolation


        #Additional features
        n=(n+1)%(m+1)
        c1=interpolation(n/m)
        pt1=Point(*c1,radius=0.1,color=colors.lighten(colors.RED,2),fill=True)
        c2=interpolation.trajectory(n/m)
        pt2=Point(*c2,radius=0.1,color=colors.lighten(colors.GREEN,2),fill=True)

        #Show visual components
        pt1.show(context)
        pt2.show(context)
        interpolation.show(context,200)
        interpolation.showTrajectory(context)

        #Console
        context.showConsole()

        #Flip the context
        context.flip()



import random


class Rect:
    """Define a pure and simple rectangle."""

    # Class methods
    @classmethod
    def cross(cls, r1, r2):
        """Determine the rectangle resulting of the intersection of two rectangles."""
        if r1.xmax < r2.xmin or r1.xmin > r2.xmax: return
        if r1.ymax < r2.ymin or r1.ymin > r2.ymax: return
        xmin = max(r1.xmin, r2.xmin)
        ymin = max(r1.ymin, r2.ymin)
        xmax = min(r1.xmax, r2.xmax)
        ymax = min(r1.ymax, r2.ymax)
        return Rect.createFromCorners(xmin, ymin, xmax, ymax)

    @classmethod
    def random(cls, borns=[-1, 1], borns_size=[0, 1]):
        """Create a random rect."""
        x = random.uniform(*borns)
        y = random.uniform(*borns)
        sx = random.uniform(*borns_size)
        sy = random.uniform(*borns_size)
        return cls(x, y, sx, sy)

    @classmethod
    def createFromCorners(cls, *corners):
        """Create a rectangle."""
        x, y, xm, ym = corners
        w = xm - x
        h = ym - y
        return cls(x + w / 2, y + h / 2, w, h)

    @classmethod
    def createFromCoordinates(cls, *coordinates):
        """Create a rect using the coordinates."""
        return cls(*coordinates)

    @classmethod
    def createFromRect(cls, *rect):
        """Create a rect from an unpacked pygame.rect"""
        l, r, w, h = rect
        return cls(l + w / 2, r + h / 2, w, h)

    def __init__(self, x, y, w, h):
        """Create a rectangle using its x, y, width, and height, the
        x and y components correspond to the center of the rectangle."""
        self.components = [x, y, w, h]

    def __getitem__(self, index):
        return self.components[index]

    def __setitem__(self, key, value):
        self.components[key] = value

    x = property(lambda cls: cls.__getitem__(0), lambda cls, value: cls.__setitem__(0, value),
                 doc="x component of the center")
    y = property(lambda cls: cls.__getitem__(1), lambda cls, value: cls.__setitem__(1, value),
                 doc="y component of the center")

    def getSize(self):
        return [self.w, self.h]

    def setSize(self, size):
        self.w, self.height = size

    size = property(getSize, setSize)

    def getPosition(self):
        return [self.x, self.y]

    def setPosition(self, position):
        self.x, self.y = position

    center = position = property(getPosition, setPosition)

    def __str__(self, n=2):
        """Return the string representation of a rect."""
        r = self.__round__(n)
        return "Rect(x=" + str(r.x) + ",y=" + str(r.y) + ",w=" + str(r.w) + ",h=" + str(r.h) + ")"

    def __round__(self, n=2):
        """Round the components of the rect."""
        x = round(self.x, n)
        y = round(self.y, n)
        w = round(self.w, n)
        h = round(self.h, n)
        return Rect(x, y, w, h)

    def __contains__(self, position):
        """Determine if a position is in the rectangle."""
        x, y = position
        return (self.xmin <= x <= self.xmax) and (self.ymin <= y <= self.ymax)

    def resize(self, n):
        """Allow the user to resize the rectangle."""
        self.w *= n
        self.h *= n

    def __iter__(self):
        self.iterator = 0
        return self

    def __next__(self):
        if self.iterator < 4:
            self.iterator += 1
            return self.components[self.iterator - 1]
        else:
            raise StopIteration

    # properties
    # corners
    def getCorners(self):
        """Return the corners of the rect."""
        return [self.x - self.w / 2, self.y - self.h / 2, self.x + self.w / 2, self.y + self.h / 2]

    def setCorners(self, corners):
        """Set the corners of the rect."""
        x1, y1, x2, y2 = corners
        self.w = x2 - x1
        self.h = y2 - y1
        self.x = x1 - self.w / 2
        self.y = y1 - self.w / 2

    # coordinates
    def getCoordinates(self):
        """Return the coordinates of the rect."""
        return [self.x, self.y, self.w, self.h]

    def setCoordinates(self, coordinates):
        """Set the coordinates of the rect."""
        self.position = coordinates[:2]
        self.size = coordinates[2:]

    # rect
    def getRect(self):
        """Return the rect of the rectangle."""
        return Rect.getRectFromCoordinates(self.getCoordinates())

    def setRect(self, rect):
        """Set the rect of the rectangle."""
        self.setCoordinates(Rect.getCoordinatesFromRect(rect))

    # sx component
    def getWidth(self):
        """Return the width."""
        return self.components[2]

    def setWidth(self, w):
        """Set the width."""
        self.components[2] = w

    # sy component
    def getHeight(self):
        """Return the height."""
        return self.components[3]

    def setHeight(self, h):
        """Set the height."""
        self.components[3] = h

    # xmin component
    def getXmin(self):
        """Return the minimum of the x component."""
        return self.x - self.w / 2

    def setXmin(self, xmin):
        """Set the minimum of the x component."""
        self.x = xmin + w / 2

    # ymin component
    def getYmin(self):
        """Return the minimum of the y component."""
        return self.y - self.h / 2

    def setYmin(self, ymin):
        """Set the minimum of the y component."""
        self.y = ymin + self.h / 2

    # xmax component
    def getXmax(self):
        """Return the maximum of the x component."""
        return self.x + self.w / 2

    def setXmax(self, xmax):
        """Set the maximum of the x component."""
        self.x = xmax - self.w / 2

    # ymax component
    def getYmax(self):
        """Return the maximum of the y component."""
        return self.y + self.h / 2

    def setYmax(self, ymax):
        """Set the maximum of the y component."""
        self.y = ymax - self.h / 2

    corners = property(getCorners, setCorners, doc="Corners")
    coordinates = property(getCoordinates, setCoordinates, doc="Center+Size")
    w = sx = width = property(getWidth, setWidth, doc="Width")
    h = sy = height = property(getHeight, setHeight, doc="Height")
    xmin = x1 = left = l = property(getXmin, setXmin, doc="Left")
    xmax = x2 = right = r = property(getXmax, setXmax, doc="Right")
    ymin = y1 = bottom = b = property(getYmin, setYmin, doc="Bottom")
    ymax = y2 = top = t = property(getYmax, setYmax, doc="Top")

    # Static methods
    @staticmethod
    def getCornersFromCoordinates(coordinates):
        """Return the corners (top_left_corner,bottom_right_corner) using the coordinates (position+size)."""
        """[x,y,sx,sy] -> [mx,my,Mx,My]"""
        x, y, sx, sy = coordinates
        mx, my = x - sx / 2, y - sy / 2
        Mx, My = x + sx / 2, y + sy / 2
        return [mx, my, Mx, My]

    @staticmethod
    def getCoordinatesFromCorners(corners):
        """Return the coordinates (position+size) using the corners (top_left_corner,bottom_right_corner)."""
        """[mx,my,Mx,My] -> [x,y,sx,sy]"""
        mx, my, Mx, My = corners
        sx, sy = Mx - mx, My - my
        x, y = mx + sx / 2, my + sy / 2
        return [x, y, sx, sy]

    @staticmethod
    def getCoordinatesFromRect(rect):
        """Return the coordinates (position,size) using the rect (top_left_corner,size)."""
        """[x,y,sx,sy] -> [mx,my,sx,sy]"""
        mx, my, sx, sy = rect
        x, y = mx + sx / 2, my + sy / 2
        return [x, y, sx, sy]

    @staticmethod
    def getRectFromCoordinates(coordinates):
        """Return the rect (top_left_corner,size) using the coordinates (position,size)."""
        """[mx,my,sx,sy] -> [x,y,sx,sy]"""
        x, y, sx, sy = coordinates
        mx, my = x - sx / 2, y - sy / 2
        return [mx, my, sx, sy]

    @staticmethod
    def getRectFromCorners(corners):
        """Return the rect (top_left_corner,size) using the corners (top_left_corner,bottom_right_corner)."""
        """[mx,my,Mx,My] -> [mx,my,sx,sy]"""
        mx, my, Mx, My = corners
        sx, sy = Mx - mx, My - my
        return [mx, my, sx, sy]

    @staticmethod
    def getCornersFromRect(rect):
        """Return the (top_left_corner,bottom_right_corner) using the corners rect (top_left_corner,size)."""
        """[mx,my,Mx,My] -> [mx,my,sx,sy]"""
        mx, my, sx, sy = rect
        Mx, My = mx + sx, my + sy
        return [mx, my, Mx, My]


if __name__ == "__main__":
    r1 = Rect.random()
    r2 = Rect.random()
    r1.x -= 1
    print(r1, r2)
    print(r1.corners)
    print(r1.coordinates)
    print(r1.x, r1.y)
    print(r1.sx, r1.sy)
    print(r1.width, r1.height)
    r = Rect.cross(r1, r2)
    print(*r1)



"""
Features:
1:COMPLETED: print #renvoie str
2:COMPLETED: addition #renvoie Polynomial
3:COMPLETED: multiplication #renvoie Polynomial
4:COMPLETED: derive #renvoie un Polynomial
5:COMPLETED: division #renvoie couple de Polynomial
6:COMPLETED: pgcd #renvoie Polynomial
7:COMPLETED: ppcm #renvoie Polynomial
8:COMPLETED: taylor #renvoie Polynomial
9: factoriser #renvoie str ou liste de liste des coefficients "pareil"
10: racines reelles #renvoie liste des racines reelles
11: racines complexes #renvoie liste des racines complexes
12: Decomposition en éléments simples
12: show #Affiche le polynome sur une fenetre pygame, pyglet ou matplotlib
"""

from copy import deepcopy
# import matplotlib.pyplot as plt #Old way to show the polynomials on screen. Not very nice though...

# from mygrapher.grapher import Grapher #Used to show the polynome on screen with the ability to move within it.

import random  # Used only to generate random polynomials.
import functools

# Changing the type of the coefficients
#from decimal import Decimal
#from numpy import float128

# p=Polynomial.createFromInterpolation(x,y)


def prod(l): return functools.reduce(lambda a, b: a * b, l)


class Polynomial:
    def null():
        """Return the null polynomial."""
        return Polynomial([0])

    def sum(ps):
        """Sum polynomials between themselves."""
        v = Polynomial([0])  # Addition neutral
        for p in ps:
            v += p
        return v

    def prod(ps):
        """Multiply polynomials between themselves."""
        v = Polynomial([1])  # Multiplication neutral
        for p in ps:
            v *= p
        return v

    def createFromInterpolation(x, y):
        """Create a polynomial using the lagrangian interpolation."""
        n = max(len(x), len(y))
        p = Polynomial.sum([Polynomial([y[j]]) * Polynomial.prod([Polynomial([-x[i] / (
            x[j] - x[i]), 1 / (x[j] - x[i])]) for i in range(n) if x[i] != x[j]]) for j in range(n)])
        return Polynomial(p.coefficients)

    def __init__(self, coefficients, definition=[-10**20, 10**20]):
        """Create a polynomial using its coefficients."""
        self.coefficients = coefficients
        self.definition = definition

    def make(self, f):
        """Transform all the coefficients into a decimal type."""
        self.coefficients = [f(c) for c in self.coefficients]
        self.definition = [f(d) for d in self.definition]

    def __str__(self):
        """Return the usual representation of a polynomial."""
        if self.coefficients == []:
            return "0"
        liste = []
        if self.coefficients[0] != 0:
            liste.append(str(self.coefficients[0]))
        for degre, coefficient in enumerate(self.coefficients[1:], 1):
            if coefficient != 0:
                liste.append(str(coefficient) + "X^" + str(degre))
        liste.reverse()
        text = "+".join(liste)
        text = text.replace("^1", "").replace("+-", "-").replace("1X", "X")
        return text

    def adapt(coefficients, size):
        """Add zeros to the coefficients using coefficients and size."""
        return coefficients + [0] * (size - len(coefficients))

    def __eq__(self, other):
        """Determine if two polynomials are equals by checking if the coefficients
        are the same."""
        return self.coefficients == other.coefficients

    def __add__(self, other):
        """Add 2 polynomials together by adding their coefficients."""
        maxdegree = max(self.degree + 1, other.degree + 1)
        cA = Polynomial.adapt(self.coefficients, maxdegree)
        cB = Polynomial.adapt(other.coefficients, maxdegree)
        coefficients = [a + b for (a, b) in zip(cA, cB)]
        A = Polynomial(coefficients)
        A.correct()
        return A

    def __sub__(self, other):
        """Substract 2 polynomials together by multiplying the second polynomial by -1 and adding it to the first using add method."""
        return self + other * Polynomial([-1])

    def __mul__(self, other):
        """Multiply 2 polynomials together using the formula for the coefficients."""
        degree = len(self.coefficients) + len(other.coefficients)
        A = self.coefficients + [0] * (degree - len(self.coefficients) + 1)
        B = other.coefficients + [0] * (degree - len(other.coefficients) + 1)
        newcoefficients = []
        for k in range(degree + 1):
            coeff = 0
            for i in range(k + 1):
                coeff += A[i] * B[k - i]
            newcoefficients.append(coeff)
        C = Polynomial(newcoefficients)
        C.correct()
        return C

    def __pow__(self, n):
        """Does an ~exponentiation of the polynomial."""
        return Polynomial.mul([self for i in range(n)])

    def correct(self):
        """Correct the coefficients of the polynomial by deleting useless zeros at the end,"""
        while len(self.coefficients) > 1:
            if self.coefficients[-1] == 0:
                del self.coefficients[-1]
            else:
                break
        if self.coefficients == [0]:
            self.coefficients = []
        """and then replace float coefficients by int coefficients if possible."""
        for i in range(len(self.coefficients)):
            if self.coefficients[i] == int(self.coefficients[i]):
                self.coefficients[i] = int(self.coefficients[i])

    def __len__(self):
        """Return the number of coefficients.
        It is important to notice that it is not the degree due to the fact that
        the polynomial might have a degree of -inf and the len method must have positive values.
        In such a case this method will raise an error.
        It is why it is better to use the degree method in the general case."""
        A = deepcopy(self)
        A.correct()
        return len(A.coefficients) - 1

    def getDegree(self):
        """Return the degree of a polynomial.
        It will return -1 if the degree is -inf to be computer friendly."""
        A = deepcopy(self)
        A.correct()
        return len(A.coefficients) - 1

    def __getitem__(self, index):
        """Return a coefficient of the polynomial."""
        return self.coefficients[index]

    def __setitem__(self, index, value):
        """Change a coefficient of the polynomial."""
        self.coefficients[index] = value

    def __delitem__(self, index):
        """Delete a coefficient of the polynomial."""
        del self.coefficients[index]

    def derivate(self):
        """Derivate the polynomial."""
        self.coefficients = [i * c for (i, c) in enumerate(self.coefficients)]
        del self.coefficients[0]

    def derivative(self):
        """Return the derivative polynomial."""
        newcoefficients = [i * c for (i, c) in enumerate(self.coefficients)]
        del newcoefficients[0]
        return Polynomial(newcoefficients)

    def unit(self):
        """Return the polynomial unit form."""
        k = self.coefficients[-1]
        newcoefficients = []
        for i in range(len(self.coefficients)):
            newcoefficients.append(self.coefficients[i] / k)
        A = Polynomial(newcoefficients)
        A.correct()
        return A

    def unitarize(self):
        """Convert the polynomial into its unit form."""
        self.coefficients = self.unit().coefficients

    def __xor__(self, other):
        """Return the HCF (PGCD  en francais) of self and other by Euclide algorithm. Make use of "^" operator."""
        R0 = deepcopy(self)
        R1 = deepcopy(other)
        R0.correct()
        R1.correct()
        listR = [R0, R1]
        while listR[-1].degree > 0:
            R, Q = listR[-2] / listR[-1]
            listR.append(R)
        return listR[-2]

        # Resultat d'une enorme galere de 2h...
        # Algorithme d'Euclide:
        # a=b*q0+r0
        # b=r*q1+r1
        # r=r1*q2+r2
        # r1=r2*q3+r3
        # ...
        # ri=r(i-1)*q(i-1)+1+0

    def __or__(self, other):
        """Return the LCM (PPCM en francais) using "i have absolutely no clue how to do this so figure it out!" Make use of "|" operator."""
        # On sent la (tres) grosse fatigue apres minuit.
        return "Figure it out!"

    def __truediv__(self, other):
        """Return the division of 2 polynomials."""
        R = deepcopy(self)
        B = deepcopy(other)
        R.correct()
        B.correct()
        Q = Polynomial([])
        while R.degree >= B.degree:
            d = R.degree - B.degree + 1
            C = Polynomial([0 for i in range(d)])
            C[-1] = R[-1] / B[-1]
            R = R - B * C
            Q = Q + C
        return [R, Q]

        # Ce que j'ai du faire pour trouver l'implementation...
        # On a: A=BQ+R
        # [3,2,4,5,1]/[1,4,1]
        # [3,2,3,1,0]=[3,2,4,5,1]-[0,0,1]*[1,4,1] #[0,0,1,4,1]
        # [3,1,-1,0]=[3,2,3,1]-[0,1]*[1,4,1] #[0,1,4,1]
        # [4,5,0]=[3,1,-1]-[-1]*[1,4,1] #[-1,-4,-1]
        # [4,5]

        # 5=8*0+5 cas particulier de la division euclidienne

    def isDivisible(self, other):
        """Determine if the polynomial is divisible by other."""
        q, r = self / other
        return r == Polynomial.zero

    def __call__(self, x):
        """Evaluate the polynomial given an x value."""
        return sum([c * x**n for (n, c) in enumerate(self.coefficients)])

    def getFactoredForm(self):
        """Return the list of the factors of the factored form of the polynomial."""
        return

    def allDerivatives(self):
        """Return the list of all derivatives of the polynome until the zero polynome."""
        B = deepcopy(self)
        derivatives = [deepcopy(self)]
        while B.degree > 1:
            B.derivate()
            derivatives.append(deepcopy(B))
        return derivatives

    def getRootsByDichotomy(self, precision=1e-15):
        """Return the list of all the roots of the polynomial."""
        """Does that by finding where all the derivates cancel, to find new extrema and applying tvi."""
        roots = []
        derivatives = self.allDerivatives()
        derivatives.reverse()
        old_extrema = []
        for derivative in derivatives:
            new_extrema = []
            neglim = self.definition[0]
            poslim = self.definition[1]
            extrema = [neglim] + old_extrema + [poslim]
            for i in range(len(extrema) - 1):
                if derivative(extrema[i]) * derivative(extrema[i + 1]) <= 0:
                    new_extrema.append(derivative.dichotomy(
                        extrema[i], extrema[i + 1], precision))
            old_extrema = new_extrema[:]
        return new_extrema

    def dichotomy(self, xa, xb, precision=1e-15):
        """Return a float value x between a and b for which the polynomial canceled itself.
        Does that based on theoreme of intermediate value with dichotomy."""
        if self(xa) * self(xb) > 0:
            raise Exception(
                "The polynomial cannot cancel itself within this interval.")
        a = min(xa, xb)
        b = max(xa, xb)
        x = (a + b) / 2
        while abs(self(x)) > precision:
            print(x, self(x), abs(self(x)), precision)
            if self(x) * self(b) > 0:
                b = x
            if self(x) * self(a) > 0:
                a = x
            x = (a + b) / 2
        return x

    def getRoots(self, precision=1e-15):
        """
        Return the roots of the polynomial by doing the best possible calculation
        that combines precision and efficiency.
        """
        degree = self.degree  # There are less calculations by selecting the degree
        if degree == 0:
            if self.coefficients[0] == 0:
                raise Exception("Every number is a root of p.")
            else:
                return []
        elif self.degree == 1:
            b, a = self.coefficients
            return [-b / a]  # a cannot be null
        elif self.degree == 2:
            c, b, a = self.coefficients
            discriminant = b**2 - 4 * a * c
            x1 = (-b - discriminant**(1 / 2)) / (2 * a)  # a cannot be null
            x2 = (-b + discriminant**(1 / 2)) / (2 * a)  # a cannot be null
            return [x1, x2]
        else:
            return self.getRootsByDichotomy(precision)

    def getMultiplicity(self, root):
        """Return the multiplicity of a root."""
        b = self
        n = 0
        while b.coefficients != [1]:
            q, b = b / Polynomial([-root, 1])
            n += 1
        return n

    def getRootsWithMultiplicity(self):
        """Return a list of couples of roots with their multiplicity associated.
        Example: [(r1,m1),(r2,m2),...] """
        return [(r, self.getMultiplicity(r)) for r in list(set(self.roots))]

    def __floordiv__(self, other):
        """Return a rational function of numerator the polynomial, and the denominator the other."""
        return RationalFunction(self, other)

    def show(self):
        """Show the polynomial with grapher."""
        Grapher([self])()  # Instance Grapher only takes the list of the functions in parameter and can be called to display them.

    def matplotlibShow(self, zone=[-5, 5], precision=0.1):
        """Show the polynomial with matplotlib using optional zone and precision."""
        z, Z = zone
        X = list(
            [precision * x for x in range(int(z / precision), int(Z / precision))])
        Y = list([self(x) for x in X])
        plt.plot(X, Y, label="Polynomial")
        plt.show()

    degree = property(getDegree)
    roots = property(getRoots)


class FactoredPolynomial(Polynomial):
    """For factored of a polynomial."""

    def __init__(self, roots_and_multiplicity):
        """Define the factored polynomial with its roots."""
        self.roots = roots

    def __str__(self):
        """Return a string representation of the factored polynomial."""

    def getCoefficients(self):
        """Return the coefficients of the polynomial."""


class RationalFunction:
    """Wikipedia: In algebra, the partial fraction decomposition or partial
    fraction expansion of a rational function (that is, a fraction such that the
    numerator and the denominator are both polynomials) is an operation that
    consists of expressing the fraction as a sum of a polynomial (possibly zero)
    and one or several fractions with a simpler denominator"""

    def __init__(self, p, q):
        """Define a rational function using two polynomials p: the numerator
        and q: the denominator."""
        self.p = p
        self.q = q

    def __str__(self):
        """String representation of a rational function."""
        return "(" + str(self.p) + ")/(" + str(self.q) + ")"

    def __call__(self, x):
        """Evaluate in x the rational function."""
        return self.p(x) / self.q(x)

    def __add__(self, other):
        """Add two rational functions."""
        p = self.p * other.q + other.p * self.q
        q = self.q * other.q
        return RationalFunction(p, q)

    def __sub__(self, other):
        """Substract two rational functions."""
        p = self.p * other.q - other.p * self.q
        q = self.q * other.q
        return RationalFunction(p, q)

    def __mul__(self, other):
        """Multiply two rational functions."""
        p = self.p * other.p
        q = self.q * other.q
        return RationalFunction(p, q)

    def __invert__(self, other):
        """Return the inverse of the rational function. (This operation is optional.)"""
        return RationalFunction(self.q, self.p)

    def __truediv__(self, other):
        """Divide a rational function by another."""
        p = self.p * other.q
        q = self.q * other.p
        return RationalFunction(p, q)

    def getDegree(self):
        """Return the degree of the rational function."""
        return self.p.degree - self.q.degree

    def getPoles(self):
        """Return the poles of the rational function."""
        return self.q.roots

    def getRoots(self):
        """Return the roots of the rational function."""
        return self.p.roots

    def decompose(self):  # Not implemented
        """Return the partial fraction decomposition."""
        # Note this method is supposed to work with complexes
        # So it is not always valid and needs a more general case
        e, r = self.p / self.q
        roots = self.q.getRoots()
        l = [e]
        for (j, xj) in enumerate(roots):
            p = Polynomial([self.p(xj) / self.q.derivative(xj)])
            q = Polynomial([-xj, 1])
            fp = RationalFunction(p, q)
            l.append(fp)
        return l

    def getWholePart(self):  # Not finished
        """Return the whole part of the fraction."""
        e, r = self.p / self.q
        return e

    def getFractionalPart(self):  # Not finished
        """Return the fractional part of the fraction."""
        e, r = self.p / self.q
        return r

    degree = property(getDegree)
    poles = property(getPoles)


class PartialFraction(RationalFunction):  # This is just an idea
    """The only difference with the rational function is that the denominator is
    of the form: [a,b] which corresponds to q=(x-a)**b."""
    # It might not be a good idea.

    def __init__(self, p, q):
        """Define a rational function using two polynomials p: the numerator
        and q: the denominator."""
        self.p = p
        self.q = q


if __name__ == "__main__":
    # suite_random = [random.randint(0,10) for i in range(10)] #Create polynome as big as needed.
    # X=Polynomial(suite_random)
    # t=float128
    suite1 = [3, 2, 4, 5, 1]
    suite2 = [1, -2, 1]
    suite3 = [4, 5]
    A = Polynomial(suite1)
    B = Polynomial(suite2)
    C = Polynomial(suite3)
    FP = RationalFunction(A, B)
    print("A =", A)
    print("B =", B)
    print("C =", C)
    print("A+B =", A + B)
    print("A*B =", A * B)
    print("Derivative of A =", A.derivative())
    print("A/B :", A / B)
    print("A//B:", A // B)
    print("Unit polynomial of C =", C.unit())
    print("HCF (PGCD en francais) of A and B (written A^B) =", A ^ B)
    print("LCM (PPCM en francais) of A and B (written A|B) =", A | B)
    print("Roots of B:", B.roots)
    print("Multiplicity of the root 2 of B:", B.getMultiplicity(2))
    print("Roots with multiplicities of B:", B.getRootsWithMultiplicity())
    #print("Show A on screen:")
    # C=Polynomial.createFromInterpolation([1,2,-4,-6,4],[1,-3,-2,5,-2])
    print(C.coefficients)
    print(FP.decompose())
    # C.show()



from .abstract import Point,Segment,Form,Line

import random
from . import colors
import itertools

class Trajectory:
    @classmethod
    def random(cls,n=5,**kwargs):
        """Create a random trajectory of n points."""
        points=[Point.random() for i in range(n)]
        return cls(points,**kwargs)

    @classmethod
    def createFromTuples(cls,tuples,**kwargs):
        """Create a trajectory using tuples and optional arguments."""
        pts=[Point(*t) for t in tuples]
        return cls(pts,**kwargs)

    def __init__(self,points,segment_color=colors.WHITE,point_color=colors.WHITE):
        """Create a trajectory using the list of points."""
        self.points=points
        self.segment_color=segment_color
        self.point_color=point_color

    def __call__(self,t):
        """Evaluate the trajectory by t between 0 and 1."""
        if len(self.points)<=1:
            raise Exception("It is impossible to interpolate without at least 2 points.")
        length=self.length
        ns=len(self.segments)
        tsl=0
        for i in range(ns):
            s=Segment(self.points[i],self.points[i+1])
            if t*length-tsl<=s.length:
                st=(t*length-tsl)/s.length
                return tuple(s(st).components)
            tsl+=s.length
        return tuple(s(1)) #Its a fix that isn't always true

    def __str__(self):
        """Return the string representation of a trajectory."""
        return "tj("+",".join(map(str,self.points))+")"

    def show(self,surface,**kwargs):
        """Show the trajectory on the surface."""
        self.showPoints(surface,**kwargs)
        self.showSegments(surface,**kwargs)

    def showPoints(self,surface,color=None):
        """Show the point on the surface."""
        if not color: color=self.point_color
        for point in self.points:
            point.show(surface,color=color)

    def showSegments(self,surface,color=None):
        """Show the segments on the surface."""
        if not color: color=self.segment_color
        for segment in self.segments:
            segment.show(surface,color=color)

    def getSegments(self):
        """Return the segments of the trajectory of the curve."""
        return [Segment(self.points[i],self.points[i+1]) for i in range(len(self.points)-1)]

    def getLength(self):
        """Return the length of the trajectory."""
        return sum([segment.length for segment in self.segments])

    def sample(self,n,include=True):
        """Sample n points of the trajectory at equal distance.
        It is also possible to include the last one if wanted."""
        return [self(i/n) for i in range(n+int(include))]

    def fastSample(self,n,include=True):
        """Sample n points of the trajectory at equal distance but fast.
        It is also possible to include the last one if wanted."""
        length=self.length
        ns=len(self.segments)
        k=0
        l=[]
        tsl=0
        for i in range(ns):
            s=Segment(self.points[i],self.points[i+1])
            if t*length-tsl<=s.length:
                st=(t*length-tsl)/s.length
                l.append(tuple(s(st).components))
                k+=1
                t=k/n
            tsl+=s.length
        if include:
            l.append(self.points[-1])
        return l

    def sampleSegments(self,n,include=True):
        """Sample n points for each segment of the trajectory."""
        points=[]
        segments=self.segments
        for i in range(len(segments)):
            points+=segments[i].sample(n,include=False)
        if include and len(self.points)>1: points.append(self.points[-1])
        return points

    def getCenter(self):
        """Return the center (or middle) of the trajectory."""
        return self(1/2)

    def setCenter(self):
        """Set the center (or middle) of the trajectory."""
        center=self.center
        for n in range(len(self.points)):
            self.points[n]-=center

    segments=property(getSegments)
    length=property(getLength)
    center=property(getCenter)

class BezierCurve:
    def __init__(self,points,point_color=colors.WHITE,segment_color=colors.WHITE):
        """Create a bezier curve object using 3 points."""
        self.points=points
        self.segment_color=segment_color
        self.point_color=point_color

    def __call__(self,t):
        """Return the point."""
        points=self.points
        if len(points)==0:
            return None
        elif len(points)==1:
            return points[0]
        elif len(points)==2:
            segment=Segment(points[0],points[1])
            point=segment.center()
            return point
        elif len(points)==3:
            p1,p2,p3=points
            s1=Segment(p1,p2)
            s2=Segment(p2,p3)
            ps1=s1(t)
            ps2=s2(t)
            s=Segment(ps1,ps2)
            return s(t)
        else:
            while len(points)>3:
                segments=[Segment(points[i],points[i+1]) for i in range(len(points)-1)]
                points=[segment(t) for segment in segments]
            b=BezierCurve(points)
            return b(t)

    def show(self,surface,p=50):
        """Show the bezier curve on the surface."""
        points=[self(i/p) for i in range(p+1)]
        segments=[Segment(points[i],points[i+1]) for i in range(p)]
        self.showPoints(surface,points)
        self.showSegments(surface,segments)

    def getConstruction(self,t):
        """Return the construction segments of the form."""
        points=self.points
        construction=[]
        while len(points)>=2:
            segments=[Segment(points[i],points[i+1]) for i in range(len(points)-1)]
            points=[segment(t) for segment in segments]
            construction.append(segments)
        return construction

    def showConstruction(self,surface,t):
        """Show the construction of the form at the t position."""
        construction=self.getConstruction(t)
        l=len(construction)
        for i in range(l):
            k=255*(i+1)/l
            color=(0,0,k)
            for segment in construction[i]:
                #print(segment)
                segment.show(surface,color=color,width=2)

    def showPoints(self,surface,points,color=None):
        """Show the points on the surface."""
        if not color: color=self.point_color
        for point in points:
            point.show(surface,color=color,radius=0.01)

    def showSegments(self,surface,segments,color=None):
        """Show the segments on the surface."""
        if not color: color=self.segment_color
        for segment in segments:
            segment.show(surface,color=color,width=2)

    def getSegments(self):
        """Return the segments of the trajectory of the curve."""
        return [Segment(points[i],points[i+1]) for i in range(len(self.points)-1)]

    segments=property(getSegments)

class BezierForm(Form):
    def show(self,context):
        """Show the curved form on the screen."""
        self.bezier.show(context)

    @property
    def bezier(self):
        """Return the bezier curve of the form."""
        return BezierCurve(self.points+[self.points[0]])



class Arrow(BezierCurve):
    """Join 2 objects with an arrow."""
    def __init__(self,points,point_color=colors.WHITE,segment_color=colors.WHITE,vector_color=colors.WHITE):
        super().__init__(self,points,point_color,segment_color,vector_color)

    def show(self,surface,p=50):
        """Show the arrow on the screen."""
        points=[self(i/p) for i in range(p)] #p=p+1 ?
        segments=[Segment(points[i],points[i+1]) for i in range(p-1)]
        vectors=[Vector(points[i],points[i+1]) for i in range(0,p-1,5)]
        self.showPoints(surface,points)
        self.showSegments(surface,segments)
        self.showVectors(surface,vectors)

    def showVectors(self,surface,vectors,color=None):
        """Show the vectors on the surface."""
        if not color: color=self.vector_color
        for vector in vectors:
            vector.show(surface,color=color)




if __name__=="__main__":
    from mycontext import Surface
    surface=Surface(name="Curves demonstration")
    l=10
    points=[Point(2*x,random.randint(-5,5)) for x in range(l)]
    t=Trajectory(points,segment_color=colors.GREEN)
    b=BezierCurve(points,segment_color=colors.RED)
    #st=t.sampleSegments(3)
    #t1=Trajectory(st,segment_color=colors.BLUE)
    #print(b.points[-1])
    #print(b(1))
    #print(b.segments[-1])
    #c=CurvedForm(points)
    n=0
    ncp=50 #number construction points

    while surface.open:
        surface.check()
        surface.control()
        surface.clear()
        surface.show()

        Point.turnPoints([1/1000 for i in range(l)],points)
        n=(n+1)%(ncp+1)
        b.showConstruction(surface,n/ncp)

        t.show(surface)
        b.show(surface)

        p1=b(n/ncp)
        p2=Point(*t(n/ncp))

        #l1=Line.createFromTwoPoints(p1,p2)

        p1.show(surface,color=colors.YELLOW,radius=0.1,fill=True)
        p2.show(surface,color=colors.YELLOW,radius=0.1,fill=True)
        #l1.show(surface,color=colors.LIGHTGREY)


        surface.flip()



import time

def timer(function):
    """Print the duration of a function making use of decorators."""
    def function_(*args,**kwargs):
        """Tested function."""
        ti=time.time()
        result=function(*args,**kwargs)
        tf=time.time()
        dt=tf-ti
        print("[TIMER]: "+str(function.__name__)+" took "+str(dt)+" seconds.")
        return result
    return function_



from draw import Draw
from rect import Rect
from window import Window

import colors
import time as tm
import copy

# For the camera only
from pygame.locals import *
import numpy as np
import pygame
import cv2


# For the camera only


class Camera:
    """The camera relies on opencv, pygame and numpy."""

    def __init__(self, draw, position=[0, 0],
                 build_capture=False,
                 build_screen_writer=False,
                 build_capture_writer=False,
                 filename="unnamed.mp4",
                 framerate=15):
        """Create an opencv camera. By default the capture is not active, it must
        be built using the build method, or by setting build to True in the
        init method."""
        self._draw = draw
        self.position = position
        self.filename = filename
        self.fourccs = {'mp4': 'MP4V', 'avi': 'DIVX'}
        self.framerate = framerate
        # Building videos components
        if build_capture:
            self.buildCapture()
        if build_screen_writer:
            self.buildScreenWriter()
        if build_capture_writer:
            self.buildCaptureWriter()

    def buildCapture(self):
        """Build the capture with opencv. Having a separate function for the
        capture allows to have more efficiency for programs that do not need the
        camera."""
        self.capture = cv2.VideoCapture(0)

    def buildCaptureWriter(self, filename=None, framerate=None):
        """Write the videos of the camera, this is just an odd concept, nothing
        really functional."""
        if not filename: filename = self.filename
        _, extension = filename.split('.')
        fourcc = self.fourccs[extension]
        print(fourcc)
        fourcc = cv2.VideoWriter_fourcc(*fourcc)
        if not framerate: framerate = self.framerate
        self.capture_writer = cv2.VideoWriter(filename, fourcc, framerate, frameSize=self._draw.window.size)

    def buildScreenWriter(self, filename=None, framerate=None):
        """Write the videos of the screen."""
        if not filename: filename = self.filename
        _, extension = filename.split('.')
        fourcc = self.fourccs[extension]
        fourcc = cv2.VideoWriter_fourcc(*fourcc)
        if not framerate: framerate = self.framerate
        self.screen_writer = cv2.VideoWriter(filename, fourcc, framerate, frameSize=self._draw.window.size)

    def writeCapture(self):
        """Write the capture."""
        # Note this cannot work if the video capture is not built
        _, frame = self.capture.read()
        self.capture_writer.write(frame)

    def writeScreen(self):
        """Write the screen."""
        output = pygame.surfarray.array3d(self._draw.window.screen)
        output = cv2.transpose(output)
        output = cv2.cvtColor(output, cv2.COLOR_RGB2BGR)
        self.screen_writer.write(output)

    def show(self):
        """Show the video."""
        ret, frame = self.capture.read()
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        frame = np.rot90(frame)
        frame = pygame.surfarray.make_surface(frame)
        self._draw.image(self._draw.window.screen, frame, self.position)

    __call__ = show

    def endCaptureWriter(self):
        """End the video writer and save its content."""
        del self.capture_writer

    def endScreenWriter(self):
        """End the screen writer and record its content."""
        del self.screen_writer

    def endCapture(self):
        """End the camera by ending the video capture."""
        del self.capture

    def destroy(self):
        """Destroy all cv2 windows."""
        cv2.destroyAllWindows()

    def isCapturing(self):
        """Determine if the capture is set."""
        return "capture" in self.__dict__

    def isWriting(self):
        """Determine if some content is being recorded."""
        return ("screen_writer" in self.__dict__) or ("capture_writer" in self.__dict__)

    def isScreenWriting(self):
        """Determine if the screen is being recorded."""
        return "screen_writer" in self.__dict__

    def isCaptureWriting(self):
        """Determine if the capture is being recorded."""
        return "capture_writer" in self.__dict__

    def setCapturing(self, capturing):
        """Set the capturing to the boolean 'capturing'."""
        if not self.capturing and capturing:
            self.buildCapture()
        if self.capturing and not capturing:
            self.endCapture()

    def setScreenWriting(self, screen_writing):
        """Set the screen writing to the boolean 'screen_writing'."""
        if not self.screen_writing and screen_writing:
            self.buildScreenWriter()
        if self.screen_writing and not screen_writing:
            self.endScreenWriter()

    def setCaptureWriting(self, capture_writing):
        """Set the capture writing to the boolean 'capture_writing'."""
        if not self.capture_writing and capture_writing:
            self.buildCaptureWriter()
        if self.capture_writing and not capture_writing:
            self.endCaptureWriter()

    capturing = property(isCapturing, setCapturing)
    screen_writing = property(isScreenWriting)
    capture_writing = property(isCaptureWriting)
    writing = property(isWriting)

    def switchCapture(self):
        """Switch the capture mode."""
        if self.capturing:
            self.endCapture()
        else:
            self.buildCapture()

    def switchScreenWriting(self):
        """Switch the screen writing mode."""
        if self.screen_writing:
            self.endScreenWriter()
        else:
            self.buildScreenWriter()

    def switchCaptureWriting(self):
        """Switch the capture writing mode."""
        if self.capture_writing:
            self.endCaptureWriter()
        else:
            self.buildCaptureWriter()

    def __del__(self):
        """Delete the camera components if they exist."""
        if self.capturing:
            del self.capture
        if self.screen_writing:
            del self.screen_writer
        if self.capture_writing:
            del self.capture_writer
        self.destroy()

    def write(self):
        """Write the screen or the capture video if they are on."""
        if self.screen_writing:
            self.writeScreen()
        if self.capture_writing:
            self.writeCapture()

    def release(self):
        """Release the screen or the capture video if they are on."""
        if self.screen_writing:
            self.screen_writer.release()
        if self.capture_writing:
            self.capture_writer.release()


class Line:
    """Representation of a line in the console.
    This might be used for visual debugging but also for typing commands."""

    def __init__(self, *content, time=None, color=colors.WHITE, separator=" "):
        """Object of line created using the text and optional time."""
        if time is None: time=tm.time()
        self.content = list(content)
        self.time = time
        self.color = color
        self.separator = separator

    def __str__(self):
        """Return the string representation of a line."""
        return self.separator.join(map(str, self.content))

    def __iter__(self):
        """Iterate a line by giving its content of the instant the line was written."""
        self.iterator = 0
        return self

    def __next__(self):
        """Return first the content, then the instant the line was written."""
        if self.iterator < 2:
            self.iterator += 1
            if self.iterator + 1 == 0:
                return self.content
            elif self.iterator + 1 == 1:
                return self.time

    def getText(self):
        """Return the content of a line."""
        return self.content

    text = property(getText)

    def refresh(self):
        """Refresh the time."""
        self.time = tm.time()

    @property
    def empty(self):
        return self.content[0] == ""  # difficult to do uglier


class Console:
    def __init__(self, draw,
                 lines=[],
                 interline=15,
                 left_padding=10,
                 down_padding=20,
                 conversion=False,
                 size=20,
                 position=[0, 0],
                 font=None,
                 max_lines_shown=10,
                 duration_lines_shown=10,
                 disappearance_lines_shown=10,
                 colors=[colors.WHITE, colors.BLACK]):
        """Create a console."""
        self._draw = draw  # Draw is protected and can only be read
        self.lines = lines  # List of Lines
        self.interline = interline
        self.left_padding = left_padding
        self.down_padding = down_padding
        self.conversion = conversion
        self.size = size
        self.position = position
        self.font = font
        self.max_lines_shown = max_lines_shown
        self.duration_lines_shown = duration_lines_shown
        self.disappearance_lines_shown = disappearance_lines_shown
        self.colors = colors
        self.line_index = 0
        self.line_memory = None

    def eval(self):
        """Execute the last line."""
        content = self.lines[-1].content
        if len(content) >= 1:
            if content[0] == "marc":
                content[0] += " is awesome."
            elif content[0] == "time":
                content[0] = "The time is " + time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime()) + "."
            elif content[0] == "minecraft":
                content[0] += "is the best game ever."
            elif content[0][:6] == "print(":
                self(eval(content[0][6:-1]))
            elif content[0] == "refresh":
                self.refresh()

    def back(self):
        """Set the line to an older one."""
        if self.line_index == 0:
            self.line_memory = copy.deepcopy(self.line)
        if self.line_index < len(self.lines) - 1:
            self.line_index += 1
            self.line = copy.deepcopy(self.lines[-1 - self.line_index])
            self.line.refresh()

    def forward(self):
        """Set the line to a new one."""
        if self.line_index > 1:
            self.line_index -= 1
            self.line = copy.deepcopy(self.lines[-1 - self.line_index])
        elif self.line_index == 1:
            self.line_index -= 1
            self.line = self.line_memory
        self.line.refresh()

    def clear(self):
        """Clear the console by removing all the lines."""
        self.lines = []

    def refresh(self):
        """Refresh the console by showing its previous messages."""
        for line in self.lines:
            line.time = tm.time()

    def __getitem__(self, i):
        """Return the i-th line."""
        return self.lines[i]

    def __setitem__(self, v, i):
        """Set the i-th line."""
        self.lines[i] = v

    def __iadd__(self, *args):
        """Add a line to the console."""
        l = Line(*args, time=tm.time())
        self.lines.append(l)

    append = __iadd__

    def appendLines(self, lines):
        """Append multiple lines."""
        for line in lines:
            self.append(line)

    def __call__(self, *args, **kwargs):
        """Display a message on the context as a console would do."""
        if len(args) > 0:
            l = Line(*args, **kwargs)
            self.lines.append(l)

    def nextArg(self):
        """Create a new argument."""
        self.lines[-1].content.append("")

    def nextLine(self):
        """Create a new line."""
        self("")

    def setLine(self, line):
        self.lines[-1] = line

    def getLine(self):
        return self.lines[-1]

    def delLine(self):
        del self.lines[-1]

    line = property(getLine, setLine, delLine)

    def show(self):
        """Show the console without adding a text."""
        sx, sy = self._draw.window.size
        n = len(self.lines)
        nmax = self.max_lines_shown
        dls = self.disappearance_lines_shown
        for i in range(min(n, nmax)):
            position = (self.left_padding, sy - self.interline * i - self.down_padding)
            to = self.lines[-i - 1].time
            t = (tm.time() - to) / self.duration_lines_shown
            if t < 1:
                c1 = self.lines[-i-1].color + (t ** dls,)
                # c2 = self.colors[1] + (t ** dls,)
                self._draw.print(self.lines[-i - 1], position, self.size, c1, self.font, self.conversion)


class Context(Rect):
    """The context is an object that allows the user to display graphical objects
    on the screen in a virtual mathematical plane."""

    def createFromSizeAndCorners(size, corners, **kwargs):
        c = Context(size=size, **kwargs)
        c.corners = corners
        return c

    def __init__(self, draw=None, console=None, camera=None, **kwargs):
        """Create a context using the optional draw, console and camera, with
        some of the draw's arguments for the window.
        If those are not given, they will be created automatically."""
        if draw is None:
            draw = Draw(**kwargs)
        if console is None:
            console = Console(draw)
        if camera is None:
            camera = Camera(draw)
        self.start_time = tm.time()
        self.draw = draw
        self.console = console
        self.camera = camera
        self.clear = self.draw.clear
        self.flip = self.draw.window.flip
        self.press = self.draw.window.press
        self.build = self.draw.window.build
        self.click = self.draw.window.click
        self.__call__ = self.draw.window.__call__
        self.wait = self.draw.window.wait
        self.control = self.draw.control
        self.checking = self.draw.window.checking
        self.loadImage = self.draw.window.loadImage
        self.print = self.draw.print
        self.update = self.draw.window.update
        self.switch = self.draw.window.switch
        self.count = self.draw.window.count
        self.alert = self.draw.window.alert
        self.scale = self.draw.window.scale

    def getScreen(self):
        return self.draw.window.screen

    def setScreen(self, screen):
        self.draw.window.screen = screen

    screen = property(getScreen, setScreen)

    def getKeys(self):
        """Return the keys of the window."""
        return self.press()

    def getCorners(self):
        """Return the corners of the surface shown."""
        return self.draw.plane.getCorners(self.draw.window)

    def setCorners(self, corners):
        """Set the corners of the plane using the new corners."""
        self.draw.plane.setCorners(corners, self.draw.window)

    def getRect(self):
        """Return the rect of the surface shown."""
        return Plane.getRectFromCorners(self.getCorners())

    def setRect(self, rect):
        """Set the rect of the surface shown."""
        self.draw.plane.setCorners(Plane.getCornersFromRect(rect))

    def getCoordonnates(self):
        """Return the coordonnates of the surface shown."""
        return Plane.getCoordonnatesFromCorners(self.getCorners())

    def setCoordonnates(self, coordonnates):
        """Set the coordonnates of the surface shown."""
        self.draw.plane.setCorners(
            Plane.getCornersFromCoordonnates(coordonnates))

    def getPosition(self):
        """Return the position of the surface shown."""
        return self.draw.plane.position

    def setPosition(self, position):
        """Set the position of the surface shown."""
        self.draw.plane.position = position

    def getSize(self):
        """Return the size of the surface shown."""
        return self.draw.window.size

    def setSize(self, size):
        """Set the size of the surface shown."""
        self.draw.window.size = size

    def getUnits(self):
        """Return the units of the surface shown."""
        return self.draw.plane.units

    def setUnits(self, units):
        """Set the units of the surface shown."""
        self.draw.plane.units = units

    def point(self):
        """Adapt the position of the cursor in plane's coordonnates."""
        x, y = self.draw.window.point()
        x, y = self.draw.plane.getFromScreen([x, y], self.draw.window)
        return (x, y)

    def check(self):
        """Check if the surface is open."""
        self.draw.window.check()
        self.open = self.draw.window.open

    def __call__(self):
        """Calling the surface allow the user to move on screen."""
        while self.open:
            self.check()
            self.draw.plane.control(self.draw.window)
            self.draw.window.clear()
            self.draw.plane.show(self.draw.window)
            self.flip()

    def __contains__(self, position):
        """Determine if a position is in the context."""
        return self.draw.plane.contains(position, self.draw.window)

    def show(self):
        """Show the plane on screen."""
        self.draw.plane.show(self.draw.window)

    def refresh(self):
        """Refresh the context by clearing the screen and showing the plane."""
        self.draw.window.clear()
        self.draw.plane.show(self.draw.window)

    def controlZoom(self):
        """Control the zoom of the surface's plane."""
        self.draw.plane.controlZoom(self.draw.window)

    def controlPosition(self):
        """Control the position of the surface's plane."""
        self.draw.plane.controlPosition(self.draw.window)

    def getFromScreen(self, position):
        """Behave like the get from screen of the plan without having to put
        the window in parameter."""
        return self.draw.plane.getFromScreen(position, self.draw.window)

    def getToScreen(self, position):
        """Behave like the get to screen of the plan without having to put
        the window in parameter."""
        return self.draw.plane.getToScreen(position, self.draw.window)

    def blit(self, surface, position):
        """Blit a given surface to a given position."""
        size = surface.get_size()
        position = self.draw.plane.getToScreen(position, self.window)
        self.draw.window.screen.blit(surface, position)

    def getOpen(self):
        return self.draw.window.open

    def setOpen(self, open):
        self.draw.window.open = open

    def __enter__(self):
        """Opening the context."""
        self.check()
        self.control()
        self.clear()
        self.show()
        print("went in")
        return self

    def __exit__(self):
        """Ending the context."""
        print("going out")
        self.flip()
        if self.open:
            self.__enter__()

    def getWidth(self):
        """Return the width of the window."""
        return self.size[0]

    def setWidth(self, width):
        """Set the width of the window."""
        self.size[0] = width

    def getHeight(self):
        """Return the height of the window."""
        return self.size[1]

    def setHeight(self, height):
        """Set the height of the window."""
        self.size[1] = height

    def __bool__(self):
        """Determine if the context is opened or not."""
        return self.open

    def getText(self):
        """Text is an alias for console and can be read."""
        return self.console

    def setText(self, value):
        """Text is an alias for console and can be written."""
        self.console = value

    def showConsole(self, *args):
        """Show the console, this function is a fix for deprecated programs."""
        self.console.show(*args)

    def getCounter(self):
        """Return the window counter."""
        return self.draw.window.counter

    def setCounter(self, coutner):
        """Set the window counter."""
        self.draw.window.counter = counter

    def getPlane(self):
        """Return the draw's plane."""
        return self.draw.plane

    def setPlane(self, plane):
        """Set the draw's plane."""
        self.plane = plane

    def getWindow(self):
        """Return the draw's window."""
        return self.draw.window

    def setWindow(self, window):
        """Set the draw's window."""
        self.draw.window = window

    def getFullscreen(self):
        """Determine if the context is in fullscreen mode."""
        return self.draw.window.fullscreen

    def setFullscreen(self, fullscreen):
        """Set the fullscreen mode."""
        self.draw.window.fullscreen = fullscreen

    def getResolution(self):
        """Return the resolution of the screen."""
        return self.width / self.height

    @property
    def rate(self):
        return self.counter / (tm.time() - self.start_time)

    window = property(getWindow, setWindow)
    plane = property(getPlane, setPlane)

    corners = property(getCorners, setCorners)
    rect = property(getRect, setRect)
    coordonnates = property(getCoordonnates, setCoordonnates)  # Coordinates
    position = property(getPosition, setPosition)
    size = property(getSize, setSize)
    units = property(getUnits, setUnits)
    keys = property(getKeys)
    open = property(getOpen, setOpen)
    width = property(getWidth, setWidth)
    height = property(getHeight, setHeight)
    text = property(getText, setText)
    counter = property(getCounter, setCounter)
    fullscreen = property(getFullscreen, setFullscreen)
    resolution = property(getResolution)


Surface = Context  # Ugly fix for deprectated programs

if __name__ == "__main__":
    context = Context(name="Context test")
    context.camera.buildCapture()
    context.camera.buildScreenWriter("mycontext test.mp4")
    # context.camera.buildCaptureWriter('zzzz.mp4')
    context.console.duration_lines_shown = 2
    context.console.max_lines_shown = 40
    # context.corners=[0,0,1,1]
    print(context.corners)
    print(bool(context))
    while context:
        context.count()
        print(context.counter)
        context.check()
        context.update()
        context.control()
        context.clear()
        context.show()

        context.camera.show()
        context.camera.write()
        context.console(context.rate)

        # if context.counter > 100:
        #    context.camera.destroy()

        context.flip()
    context.camera.capture.release()
    context.camera.screen_writer.release()
    print('released')



from colors import *

from pygame.locals import *


import pygame; pygame.init()

import random
import time


class Window:
    made = 0
    draw = pygame.draw
    rect = pygame.rect
    mouse = pygame.mouse
    # blit=pygame.blit
    display = pygame.display
    transform = pygame.transform

    # Creation of the window

    def __init__(self,
                 name="Unnamed",
                 size=None,
                 text_font="monospace",
                 text_size=50,
                 text_color=WHITE,
                 background_color=BLACK,
                 fullscreen=False,
                 build=True):
        """Create a window object using name, size text_font, text_size,
        text_color, background and set."""
        Window.made += 1
        self.number = Window.made
        self.name = name
        self.text_font = text_font
        self.text_size = text_size
        self.text_color = text_color
        self.background_color = background_color
        self.fullscreen = fullscreen
        self.set()
        self.log("Window has been created.")
        if build:
            self.build(size)

    def set(self):
        """Set builtins attributs of window object."""
        self.RIGHT = 0
        self.UP = 1
        self.LEFT = 2
        self.DOWN = 3
        self.open = False
        self.screenshots_taken = 0
        self.counter = 0

    def build(self, size=None):
        """Creates apparent window."""
        self.events = pygame.event.get
        self.clock = pygame.time.Clock()
        self.info = pygame.display.Info()
        self.font = pygame.font.SysFont(self.text_font, self.text_size)
        self.setScreenMode(size)
        pygame.display.set_caption(self.name)
        if self.text_color is None:
            self.text_color = self.reverseColor(self.background_color)
        self.clear()
        self.flip()
        self.open = True

    # Screen related functions

    def __call__(self, duration=float("inf")):
        """Flip the screen and pause the window."""
        self.flip()
        self.pause(duration)

    def __bool__(self):
        """Determine if the window is open or not."""
        return self.open

    def clear(self, color=None):
        """Clear to background color."""
        if color is None:
            color = self.background_color
        self.screen.fill(color)

    def flip(self):
        """Display on the screen the image considered."""
        pygame.display.flip()

    def update(self):
        """Update the screen."""
        pygame.display.update()

    def screenshot(self, image=None, name=None):
        """Save an image using the image and the name.
        - If the image is not given, a screenshot will be made.
        - If the name is not given, the named will be unnamed."""
        if name == None: name = "unnamed"
        if image == None: pygame.image.save(self.screen, name)
        self.screenshot_taken += 1

    def rename(self, name):
        """Rename the window using name."""
        self.name = name
        pygame.display.set_caption(self.name)

    def getArray(self):
        """Return the array of the screen."""
        return pygame.PixelArray(self.screen)

    def setArray(self, array):
        """Set the surface using the array."""
        surface = pygame.PixelArray.make_surface()
        self.screen.blit(surface)

    array = property(getArray, setArray)

    # Events related functions

    def check(self):
        """Update window's state depending if close buttons are pressed."""
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.open = False
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.open = False
            if event.type == VIDEORESIZE:
                self.screen = pygame.display.set_mode((event.w, event.h), RESIZABLE)

    def checking(self, event):
        """Determines if the window must be closed."""
        if event.type == pygame.QUIT:
            self.open = False
        if event.type == KEYDOWN:
            if event.key == K_ESCAPE:
                self.open = False

    def press(self):
        """Return all keys."""
        return pygame.key.get_pressed()

    def direction(self):
        """Return keys for arrows pressed. Trigonometric orientation is used."""
        keys = pygame.key.get_pressed()
        return (keys[K_RIGHT], keys[K_UP], keys[K_LEFT], keys[K_DOWN])

    def select(self):
        """Wait for user to click on screen, then return cursor position."""
        while self.open:
            for event in pygame.event.get():
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        self.open = False
                    if event.type == KEYDOWN:
                        if event.key == K_ESCAPE:
                            self.open = False
                if event.type == MOUSEBUTTONDOWN and event.button == 1:
                    return (event.pos[0], event.pos[1])

    def point(self):
        """Return cursor position on screen."""
        return pygame.mouse.get_pos()

    def click(self):
        """Return bool value for clicking on screen."""
        return bool(pygame.mouse.get_pressed()[0])

    def press(self):
        """Return bool value for clicking on screen."""
        return pygame.key.get_pressed()

    # Time related functions

    def clock(self, tick):
        """Set a clock using tick."""
        self.clock.tick(tick)

    def pause(self, duration=float("inf")):
        """Wait for user to press the 'space'."""
        to = time.time()
        while self.open and time.time() - to < duration:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.open = False
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        self.open = False
                    if event.key == K_SPACE:
                        return

    def sleep(self, duration):
        """Sleep for a given duration."""
        time.sleep(duration)

    def wait(self, duration=0.1):
        """Wait during given time."""
        t = time.time()
        while self.open and time.time() - t < duration:
            self.check()

    def count(self, n=1):
        """Increment the counter by n, which is 1 by default."""
        self.counter += n

    # Sound related functions

    def loadSound(self, filename):
        """Load a sound. Doesn't work on mac apparently..."""
        return pygame.mixer.Sound(filename)

    def playSound(self, sound):
        """Play a sound."""
        pygame.mixer.Sound.play(sound)

    def loadMusic(self, filename):
        """"Load a music. Doesn't work on mac apparently..."""
        pygame.mixer.music.load(filename)

    def playMusic(self, music, times=1):
        """Play a music."""
        pygame.mixer.music.play(music, times)

    def pauseMusic(self, music):
        """Pause a music being played."""
        pygame.mixer.music.pause()

    def unpauseMusic(self, music):
        """Unpause a music being paused."""
        pygame.mixer.music.unpause()

    def stopMusic(self, music):
        """Stop a music being played."""
        pygame.mixer.music.stop()

    def load(self, filename, music=False):
        """Load an image or a sound."""
        x = filename.split(".")[-1]
        if x == "mp3" or x == "ogg" or x == "waw":
            if music:
                return self.loadMusic()
            else:
                return self.loadSound(filename)
        if x == "png" or x == "jpg" or x == "jpeg":
            return self.loadImage(filename)

    # Image related functions

    def blit(self, image, position):
        """Blit an image on the screen."""
        self.screen.blit(image, position)

    def loadImage(self, filename):
        """Load an image using a filename."""
        return pygame.image.load(filename)

    def scale(self, picture, size):
        """Return scaled picture using picture and size."""
        return pygame.transform.scale(picture, size)

    def load(self, filename):
        """Return picture using picture directory."""
        return pygame.image.load(filename)

    def centerSurface(self, surface):
        """Centers a surface on the screen."""
        sx, sy = self.size
        sfx, sfy = surface.get_size()
        return ((sx - sfx) // 2, (sy - sfy) // 2)

    def alert(self, *text, size=None, color=None, font=None, background=None, bold=False, italic=False):
        """Quickly display text on window."""
        if not size: size = self.text_size
        if not color: color = self.text_color
        if not font: font = self.text_font
        text = " ".join(map(str, text))
        font = pygame.font.SysFont(font, size, bold, italic)
        surface = font.render(text, 1, color, background)
        position = self.centerSurface(surface)
        self.screen.blit(surface, position)
        self.flip()

    def print(self, text, position, size=None, color=None, font=None, background=None, bold=False, italic=False):
        """Display text on screen using position, size, color and font."""
        if not size: size = self.text_size
        if not color: color = self.text_color
        if not font: font = self.text_font
        font = pygame.font.SysFont(font, size, bold, italic)
        surface = font.render(str(text), 1, color[:3], background)
        if len(color) == 4:
            a = color[3]
            surface.set_alpha(a)
            raise NotImplementedError("Transparency does not work properly in pygame")
        self.screen.blit(surface, position)

    def setScreenMode(self, size=None):
        """Set the display for the screen to the right mode using its optional size."""
        if not size:
            if self.fullscreen:
                size = [self.info.current_w, self.info.current_h]
            else:
                size = [2 * self.info.current_w // 3, 2 * self.info.current_h // 3]
        if self.fullscreen:
            self.screen = pygame.display.set_mode(size, FULLSCREEN)
        else:
            self.screen = pygame.display.set_mode(size, RESIZABLE)

    def switch(self):
        """Switch the mode into fullscreen or normal."""
        self.fullscreen = not (self.fullscreen)
        self.setScreenMode()

    # Usefull functions

    def randomColor(self, alpha=False):
        """Return random color."""
        return tuple([random.randint(0, 255) for i in range(3 + int(alpha))])

    def reverseColor(self, color):
        """Return reverse color."""
        return tuple([255 - c for c in color])

    def lighten(self, color, luminosity=80):  # View later
        """Return lightened color using color and luminosity percentage."""
        r, g, b = color
        if luminosity >= 50:
            r += (255 - r) * luminosity / 100
            g += (255 - g) * luminosity / 100
            b += (255 - b) * luminosity / 100
        else:
            r -= r * luminosity / 100
            g -= g * luminosity / 100
            b -= b * luminosity / 100
        color = r, g, b
        return color

    def colorize(self, image, color):
        """Return image colorized"""
        image = image.copy()
        image.fill((0, 0, 0, 255), None, pygame.BLEND_RGBA_MULT)
        image.fill(color[0:3] + (0,), None, pygame.BLEND_RGBA_ADD)
        return image

    def wavelengthToRGB(self, wavelength):
        """Convert wavelength to rgb color type."""
        gamma, max_intensity = 0.80, 255

        def adjust(color, factor):
            if color == 0:
                return 0
            else:
                return round(max_intensity * pow(color * factor, gamma))

        if 380 <= wavelength <= 440:
            r, g, b = -(wavelength - 440) / (440 - 380), 0, 1
        elif 440 <= wavelength <= 490:
            r, g, b = 0, (wavelength - 440) / (490 - 440), 1
        elif 490 <= wavelength <= 510:
            r, g, b = 0, 1, -(wavelength - 510) / (510 - 490)
        elif 510 <= wavelength <= 580:
            r, g, b = (wavelength - 510) / (580 - 510), 1, 0
        elif 580 <= wavelength <= 645:
            r, g, b = 1, -(wavelength - 645) / (645 - 580), 0
        elif 645 <= wavelength <= 780:
            r, g, b = 1, 0, 0
        else:
            r, g, b = 0, 0, 0
        if 380 <= wavelength <= 420:
            factor = 0.3 + 0.7 * (wavelength - 380) / (420 - 380)
        elif 420 <= wavelength <= 701:
            factor = 1
        elif 701 <= wavelength <= 780:
            factor = 0.3 + 0.7 * (780 - wavelength) / (780 - 700)
        else:
            factor = 0
        r, g, b = adjust(r, factor), adjust(g, factor), adjust(b, factor)
        return (r, g, b)

    def log(self, message):
        """Print message with window mention."""
        text = "[" + str(self.name) + "] " + str(message)
        print(text)

    # End of the game

    def quit(self):
        """Quit pygame."""
        self.log("Window has been closed.")
        pygame.quit()

    kill = quit

    # Properties

    def getSize(self):
        """Return the size of the window."""
        return self.screen.get_size()

    def getCorners(self):
        """Return the corners of the screen."""
        return (0, 0) + self.size

    def getWidth(self):
        """Return the width of the screen."""
        return self.screen.get_width()

    def getHeight(self):
        """Return the height of the screen."""
        return self.screen.get_height()

    def getCenter(self):
        """Return the center of the screen."""
        return tuple([s // 2 for s in self.size])

    size = property(getSize, doc="Size can only be read not written.")
    corners = property(getCorners, doc="Corners can only be read not written.")
    width = property(getWidth, doc="Width can only be read not written.")
    height = property(getHeight, doc="Height can only be read not written.")
    center = property(getCenter, doc="Center can only be read not written")


# class Text

# class TextWritter(Window):
# Writes text on a surface


if __name__ == "__main__":
    w = Window("Window Prototype")
    # w.save(w,"grosse fenetre")
    # w=load("grosse fenetre")
    print(w.lighten(BLUE))
    # music=w.loadMusic("The-Outlander.mp3")
    # w.playMusic(music)
    w.alert("test")
    w.pause()
    w.clear()
    w.switch()
    # w.alert(w.size)
    w.alert(w.center, w.size, background=(255, 0, 0))
    w.pause()
    w.clear()
    w.alert("je raconte de la merde juste pour avoir une longue chaine de caractere")
    w.pause()
    # w.stopMusic()



from window import Window
from pygame.locals import *

import math
import numpy as np
import colors

class Plane:
    def __init__(self,theme={},view=None):
        """Create a plane using optionals theme and view."""
        self.createTheme(theme)
        self.createView(view)

    def createTheme(self,theme={}):
        """Initializes the position and the colors for the view of the plane using optional theme."""
        if not "background"  in theme: theme["background"]  = (0,0,0)
        if not "grid color"  in theme: theme["grid color"]  = (20,20,20)
        if not "grid nscale" in theme: theme["grid nscale"] = 3
        if not "show scale"  in theme: theme["show scale"]  = False
        if not "show origin" in theme: theme["show origin"] = False
        if not "show units" in theme:  theme["show units"]  = False
        self.theme=theme

    def createView(self,view=None):
        """Initializes the position and the units for the view of the plane using optional view."""
        """View must be a list of length 2."""
        if view:
            self.position= view[0]
            self.units=    view[1]
            self.default_position= view[0][:]
            self.default_units=    view[1][:]
        else:
            self.default_position= [ 0, 0]   #position of the center of the view in the plane's coordonnates
            self.default_units=    [40,40]   #units of the conversion from window/plane
            self.position=         [ 0, 0]
            self.units=            [40,40]   #units can be handled [-10**-14,10**307] before bugging or crashing

    def __call__(self,window): #The main loop must be redefined by the client and not the functions called within it.
        """Main loop of the plane."""
        window.rename("Plane")
        while window.open:
            window.check()
            self.control(window)
            self.clear(window)
            self.show(window)
            self.flip(window)

    def contains(self,position,window):
        """Determine if a position is in the visible screen."""
        x,y=position
        xmin,ymin,xmax,ymax=self.getCorners(window)
        return (xmin<=x<=xmax) and (ymin<=y<=ymax)

    def flip(self,window):
        """Flip the plane's window."""
        window.flip()

    def clear(self,window,color=None):
        """Clear the plane."""
        if not color: color=self.theme["background"]
        window.clear(color)

    def control(self,window):
        """Control the view of the plane."""
        self.controlZoom(window)
        self.controlPosition(window)

    def controlZoom(self,window):
        """Control the zoom of the plane."""
        keys=window.press()
        ux,uy=self.units
        if keys[K_RSHIFT]:
            self.zoom([1.1,1.1])
        if keys[K_LSHIFT]:# and (ux>2 and uy>2): #The second condition is useless but prevent the user from watching errors due to too far zooming out.
            self.zoom([0.9,0.9])

    def controlPosition(self,window):
        """Control the position of the plane."""
        keys=window.press()
        if keys[K_RETURN]:
            self.units=    self.default_units[:]
            self.position= self.default_position[:]
        ux,uy=self.units
        wsx,wsy=window.size
        x,y=self.position
        k=50 #Allow the user to move in the grid a reasonable velocity.
        if keys[K_LEFT]:
            x-=wsx/ux/k
        if keys[K_RIGHT]:
            x+=wsx/ux/k
        if keys[K_DOWN]:
            y-=wsy/uy/k
        if keys[K_UP]:
            y+=wsy/uy/k
        self.position=[x,y]

    def getUnitsColor(self,scale,ascale,color=None):
        """Get the color given a x position."""
        if not color: color=self.theme["grid color"]
        dscale=scale-ascale
        return [c*(2**dscale) for c in color]


    def show(self,window):
        """Show the elements on screen using the window."""
        self.showGrids(window)

    def showAll(self):
        self.showGrids(window)
        if self.theme["show scale"]:
            window.print("scale:"+str(self.getScale(window)),(10,10),size=20)
        if self.theme["show origin"] and self.contains((0,0),window):
            self.showOrigin(window)
        if self.theme["show units"]:
            window.print("units:"+str(self.units),(10,10),size=20)
        #self.showUnits(window) #Does not work for now

    def showOrigin(self,window,color=colors.WHITE,radius=3):
        """Show the origin of the plane."""
        position=self.getToScreen([0,0],window)
        window.draw.circle(window.screen,color,position,radius,0)


    def showUnits(self,window): #Not working because of error of synchronisation between text base system and normal window base.
        """Show the unit of the grid using the window."""
        px,py=self.position
        wsx,wsy=window.size
        ux,uy=self.units
        nx=int(wsx/ux)
        ny=int(wsy/uy)
        for y in range(int(-ny+py),int(ny+py)+1,10):
            for x in range(int(-nx+px),int(nx+px)+1,10):
                X,Y=self.getToScreen([x,y],window)
                window.print(str([x,y]),[X,Y],size=20)

    def showGrids(self,window,nscale=None):
        """Show the grid using the window and optional scales."""
        ascale=self.getScale(window) #ascale like actual_scale
        for scale in self.getScales(window,nscale,ascale):
            self.showGrid(window,scale)

    def getScales(self,window,nscale=None,scale=None):
        """Return the scales ."""
        if not nscale: nscale=self.theme["grid nscale"]
        if not scale: scale=self.getScale(window)
        l=[scale-i for i in range(nscale)]
        l.reverse()
        return l

    def getScale(self,window):
        """Return the actual scale of the displayed screen."""
        ux,uy=self.units
        wsx,wsy=window.size
        uwx,uwy=wsx/ux,wsy/uy
        uw=min(uwx,uwy)
        return int(math.log(uw)/math.log(10))

    def showGrid(self,window,scale=0,ascale=None):
        """Show the grid using the window."""
        if not ascale: ascale=self.getScale(window)
        xmin,ymin,xmax,ymax=self.getCorners(window) #Get the corners of the plane
        #For each line find the begining and the end in the plane's coordonnates then convert it into screen's coordonnates.
        for x in self.arange(xmin,xmax,scale):
            color=self.getUnitsColor(scale,ascale)
            start=self.getToScreen([x,ymin],window)
            end=  self.getToScreen([x,ymax],window)
            window.draw.line(window.screen,color,start,end,1)
        #Repeat the process for the y component
        for y in self.arange(ymin,ymax,scale):
            color=self.getUnitsColor(scale,ascale)
            start=self.getToScreen([xmin,y],window)
            end=  self.getToScreen([xmax,y],window)
            window.draw.line(window.screen,color,start,end,1)

    def arange(self,min,max,scale):
        """Return the list of values that are between min and max, separated by unit."""
        min=np.around(min,-scale)
        max=np.around(max,-scale)
        return np.arange(min,max,10**scale)

    def zoom(self,zoom):
        """Allow the user to zoom into the plane."""
        for i in range(2):
            self.units[i]*=zoom[i]

    def xZoom(self,zoom):
        """Allow the user to zoom into the plane according to the x component."""
        self.units[0]*=zoom

    def yZoom(self,zoom):
        """Allow the user to zoom into the plane according to the y component."""
        self.units[1]*=zoom

    def getToScreen(self,position,window):
        """Return a screen position using a position in the plane."""
        x,y=position
        px,py=self.position
        ux,uy=self.units
        wsx,wsy=window.size
        x=int((x-px)*ux+wsx/2)
        y=int(wsy/2-(y-py)*uy)
        return [x,y]

    def getAllToScreen(self,positions,window):
        """Return the list of screen positions using the list of plane positions."""
        screen_positions=[]
        for position in positions:
            screen_positions.append(self.getToScreen(position,window))
        return screen_positions

    def getFromScreen(self,position,window):
        """Return a plane position using a position in the screen."""
        x,y=position
        px,py=self.position
        ux,uy=self.units
        wsx,wsy=window.size
        x=(x-wsx/2)/ux+px
        y=(wsy/2-y)/uy+py
        return [x,y]

    def getAllFromScreen(self,position,window):
        """Return the list of plane positions using the list of screen positions."""
        plane_positions=[]
        for position in positions:
            planes_positions.append(self.getFromScreen(position,window))
        return plane_positions

    def getCorners(self,window):
        """Return the corners of the present view."""
        wsx,wsy=window.size
        ux,uy=self.units
        x,y=self.position
        lx,ly=wsx/ux,wsy/uy
        return [x-lx/2,y-ly/2,x+lx/2,y+ly/2]

        mx,my=self.getFromScreen([0,wsy],window)
        Mx,My=self.getFromScreen([wsx,0],window)
        return (mx,my,Mx,My)

    def setCorners(self,corners,window):
        """Change the actual corners of the plane by changing its position and units."""
        xmin,ymin,xmax,ymax=corners
        wsx,wsy=window.size
        self.position=[xmin+(xmax-xmin)/2,ymin+(ymax-ymin)/2]
        self.units=[wsx/(xmax-xmin),wsy/(ymax-ymin)]

    #The following functions are static methods

    def getCornersFromCoordonnates(coordonnates):
        """Return the corners (top_left_corner,bottom_right_corner) using the coordonnates (position+size)."""
        """[x,y,sx,sy] -> [mx,my,Mx,My]"""
        x,y,sx,sy=coordonnates
        mx,my=x-sx/2,y-sy/2
        Mx,My=x+sx/2,y+sy/2
        corners=(mx,my,Mx,My)
        return corners

    def getCoordonnatesFromCorners(corners):
        """Return the coordonnates (position+size) using the corners (top_left_corner,bottom_right_corner)."""
        """[mx,my,Mx,My] -> [x,y,sx,sy]"""
        mx,my,Mx,My=corners
        sx,sy=Mx-mx,My-my
        x,y=mx+sx/2,my+sy/2
        coordonnates=(x,y,sx,sy)
        return coordonnates

    def getCoordonnatesFromRect(rect):
        """Return the coordonnates (position,size) using the rect (top_left_corner,size)."""
        """[x,y,sx,sy] -> [mx,my,sx,sy]"""
        mx,my,sx,sy=rect
        x,y=mx+sx/2,my+sy/2
        coordonnates=[x,y,sx,sy]
        return coordonnates

    def getRectFromCoordonnates(coordonnates):
        """Return the rect (top_left_corner,size) using the coordonnates (position,size)."""
        """[mx,my,sx,sy] -> [x,y,sx,sy]"""
        x,y,sx,sy=coordonnates
        mx,my=x-sx/2,y-sy/2
        rect=[mx,my,sx,sy]
        return rect

    def getRectFromCorners(corners):
        """Return the rect (top_left_corner,size) using the corners (top_left_corner,bottom_right_corner)."""
        """[mx,my,Mx,My] -> [mx,my,sx,sy]"""
        mx,my,Mx,My=corners
        sx,sy=Mx-mx,My-my
        rect=[mx,my,sx,sy]
        return rect

    def getCornersFromRect(rect):
        """Return the (top_left_corner,bottom_right_corner) using the corners rect (top_left_corner,size)."""
        """[mx,my,Mx,My] -> [mx,my,sx,sy]"""
        mx,my,sx,sy=rect
        Mx,My=mx+sx,my+sy
        corners=[mx,my,Mx,My]
        return corners



if __name__=="__main__":
    window=Window(fullscreen=False)
    theme={"background":(0,0,0),"show scale":True,"show origin":True,"grid color":(200,200,200)}
    plane=Plane(theme)
    #plane.units=[1/10**300,1/10**300]
    plane(window)
